/*!
 * inferno v0.6.0
 * (c) 2016 Dominic Gannaway
 * Released under the MPL-2.0 License.
 */
var babelHelpers = {};
babelHelpers.typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj;
};

babelHelpers.extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};

babelHelpers;

function isArray(obj) {
	return obj.constructor === Array;
}

function isNullOrUndefined(obj) {
	return obj === undefined || obj === null;
}

function isInvalidNode(obj) {
	return obj === undefined || obj === null || obj === false;
}

function isFunction(obj) {
	return typeof obj === 'function';
}

function isAttrAnEvent(attr) {
	return attr[0] === 'o' && attr[1] === 'n' && attr.length > 3;
}

/*
export function isAttrAComponentEvent(attr) {
	return attr.substring(0, 11) === 'onComponent' && attr.length > 12;
}
*/

function isAttrAHook(hook) {
	return hook === 'onCreated' || hook === 'onAttached' || hook === 'onWillDetach' || hook === 'onWillUpdate' || hook === 'onDidUpdate';
}

function isAttrAComponentHook(hook) {
	return hook === 'onComponentWillMount' || hook === 'onComponentDidMount' || hook === 'onComponentWillUnmount' || hook === 'onComponentShouldUpdate' || hook === 'onComponentWillUpdate' || hook === 'onComponentDidUpdate';
}

function createAttrsAndEvents(props, tag) {
	var events = null;
	var hooks = null;
	var attrs = null;
	var className = null;
	var style = null;

	if (!isNullOrUndefined(props)) {
		if (isArray(props)) {
			return props;
		}
		for (var prop in props) {
			if (prop === 'className') {
				className = props[prop];
			} else if (prop === 'style') {
				style = props[prop];
			} else if (isAttrAHook(prop) && !isFunction(tag)) {
				if (isNullOrUndefined(hooks)) {
					hooks = {};
				}
				hooks[prop.substring(2).toLowerCase()] = props[prop];
				delete props[prop];
			} else if (isAttrAnEvent(prop) && !isFunction(tag)) {
				if (isNullOrUndefined(events)) {
					events = {};
				}
				events[prop.substring(2).toLowerCase()] = props[prop];
				delete props[prop];
			} else if (isAttrAComponentHook(prop) && isFunction(tag)) {
				if (isNullOrUndefined(hooks)) {
					hooks = {};
				}
				hooks['c' + prop.substring(3)] = props[prop];
				delete props[prop];
			} else if (!isFunction(tag)) {
				if (isNullOrUndefined(attrs)) {
					attrs = {};
				}
				attrs[prop] = props[prop];
			} else {
				attrs = props;
			}
		}
	}
	return { attrs: attrs, events: events, className: className, style: style, hooks: hooks };
}

function createChild(_ref) {
	var tag = _ref.tag;
	var attrs = _ref.attrs;
	var children = _ref.children;
	var className = _ref.className;
	var style = _ref.style;
	var events = _ref.events;
	var hooks = _ref.hooks;

	if (tag === undefined && !isNullOrUndefined(attrs) && !attrs.tpl && !isNullOrUndefined(children) && children.length === 0) {
		return null;
	}
	var key = !isNullOrUndefined(attrs) && !isNullOrUndefined(attrs.key) ? attrs.key : null;

	if (!isNullOrUndefined(children) && children.length === 0) {
		children = null;
	} else if (!isInvalidNode(children)) {
		children = isArray(children) && children.length === 1 ? createChildren(children[0]) : createChildren(children);
	}

	if (key !== null) {
		delete attrs.key;
	}
	var attrsAndEvents = createAttrsAndEvents(attrs, tag);

	return {
		dom: null,
		tag: tag,
		key: key,
		attrs: attrsAndEvents.attrs,
		events: events || attrsAndEvents.events,
		hooks: hooks || attrsAndEvents.hooks,
		className: className || attrsAndEvents.className,
		style: style || attrsAndEvents.style,
		children: children,
		instance: null
	};
}

function createChildren(children) {
	var childrenDefined = !isNullOrUndefined(children);
	if (childrenDefined && isArray(children)) {
		var newChildren = [];

		for (var i = 0; i < children.length; i++) {
			var child = children[i];
			if (!isNullOrUndefined(child) && (typeof child === 'undefined' ? 'undefined' : babelHelpers.typeof(child)) === 'object') {
				if (isArray(child)) {
					if (child.length > 0) {
						newChildren.push(createChildren(child));
					} else {
						newChildren.push(null);
					}
				} else {
					newChildren.push(createChild(child));
				}
			} else {
				newChildren.push(child);
			}
		}
		return newChildren;
	} else if (childrenDefined && (typeof children === 'undefined' ? 'undefined' : babelHelpers.typeof(children)) === 'object') {
		return children.dom === undefined ? createChild(children) : children;
	}
	return children;
}

function createElement(tag, props) {
	for (var _len = arguments.length, children = Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
		children[_key - 2] = arguments[_key];
	}

	return createChild({ tag: tag, attrs: props, children: children });
}

function createElement$1(tag, namespace) {
	if (isNullOrUndefined(namespace)) {
		return document.createElement(tag);
	} else {
		return document.createElementNS(namespace, tag);
	}
}

// Runs only once in applications lifetime
var isBrowser = typeof window !== 'undefined' && window.document;

function createStaticElement(tag, attrs) {
	if (isBrowser) {
		var dom = createElement$1(tag);
		if (attrs) {
			createStaticAttributes(attrs, dom);
		}
		return dom;
	}
	return null;
}

function createStaticAttributes(attrs, dom) {
	var attrKeys = Object.keys(attrs);

	for (var i = 0; i < attrKeys.length; i++) {
		var attr = attrKeys[i];
		var value = attrs[attr];

		// TODO! What about SVG?
		if (attr === 'className') {
			dom.className = value;
		} else {
			if (value === true) {
				dom.setAttribute(attr, attr);
			} else if (!isNullOrUndefined(value) && value !== false && !isAttrAnEvent(attr)) {
				dom.setAttribute(attr, value);
			}
		}
	}
}

/*
function createStaticChildren(children, parentDom) {
	if (isArray(children)) {
	} else if (isStringOrNumber(children)) {
		parentDom.textContent = children;
	}
}
*/

var index = {
	createElement: createElement,
	staticCompiler: {
		createElement: createStaticElement
	}
};

export default index;